/*

sample project for using a ASM ISR for ultra efficient quadratur encoder sampling

*/

#include <avr/io.h>
#include <avr/interrupt.h>


#define XTAL        8e6         // 8MHz

#define PHASE_A     (PINC & 1<<PC1)
#define PHASE_B     (PINC & 1<<PC2)

#define LEDS_DDR    DDRD
#define LEDS        PORTD           // LEDs against VCC

volatile int8_t enc_delta;          // -128 ... 127
volatile int8_t enc_last;

void encode_init( void )
{
  int8_t enc_new;

  enc_new = 0;
  if( PHASE_A ) enc_new = 3;
  if( PHASE_B ) enc_new ^= 1;           // convert gray to binary
  enc_last = enc_new;                   // power on state
  enc_delta = 0;
  TCCR0A = (1<<WGM01) | (1<<CS01) | (1<<CS00);   // CTC, prescaler 64
  OCR0A = (uint8_t)(XTAL / 64.0 * 1e-3 - 0.5);   // 1ms
  TIMSK0 |= 1<<OCIE0A;
}


ISR( TIMER0_COMPA_vect ) {            // 1ms for manual movement

  int8_t enc_new, diff;

  enc_new = 0;
  if( PHASE_A ) enc_new = 3;
  if( PHASE_B ) enc_new ^= 1;      // convert gray to binary
  diff = enc_last - enc_new;       // difference last - new
  if( diff & 1 ) {                 // bit 0 = value (1)
    enc_last = enc_new;            // store new as next last
    enc_delta += (diff & 2) - 1;   // bit 1 = direction (+/-)
  }
}


int8_t encode_read1( void )         // read single step encoders
{
  int8_t val;

  cli();
  val = enc_delta;
  enc_delta = 0;
  sei();
  return val;                   // counts since last call
}


int8_t encode_read2( void )         // read two step encoders
{
  int8_t val;

  cli();
  val = enc_delta;
  enc_delta = val & 1;
  sei();
  return val >> 1;
}


int8_t encode_read4( void )         // read four step encoders
{
  int8_t val;

  cli();
  val = enc_delta;
  enc_delta = val & 3;
  sei();
  return val >> 2;
}


int main( void )
{
  int32_t val = 0;

  LEDS_DDR = 0xFF;
  encode_init();
  sei();

  for(;;){
    val += encode_read1();          // read a single step encoder
    LEDS = val;
  }
}
